/*******************************************************************************
    Copyright (c) 2014 NVIDIA Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be
        included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*******************************************************************************/

#ifndef _UVM_CHANNEL_MGMT_COMMON_H_
#define _UVM_CHANNEL_MGMT_COMMON_H_

#if defined (__linux__)
#include "uvm_linux.h"
#else
#include "os-kernel.h"
#include <ntddk.h>
#endif

#include "uvmtypes.h"
#include "uvm_page_migration.h"

/*
#define UVM_PUSHBUFFER_RESERVATION_SIZE         (unsigned)(64 * 4096)

#define UVM_RINGBUFFER_POOL_DEFAULT_SIZE        (16u)
#define UVM_RINGBUFFER_DEFAULT_SIZE             (unsigned)(2 * 1024 * 1024)

#define UVM_CHANNEL_POOL_DEFAULT_SIZE           (16u)
*/

// TODO: This is only for testing. Should be removed (use the above instead).
#define UVM_PUSHBUFFER_RESERVATION_SIZE         (unsigned)(512)

#define UVM_RINGBUFFER_POOL_DEFAULT_SIZE        (2u)
#define UVM_RINGBUFFER_DEFAULT_SIZE             (unsigned)(2 * 512)

#define UVM_CHANNEL_POOL_DEFAULT_SIZE           (2u)

#define UVM_READ_SEMA(sema) (*(volatile NvU32*)sema)

//
// Example usage:
//
//   NvBool isPushSuccess;
//  
//   UVM_PUSH_METHOD(
//       isPushSuccess, pushbuffer, ceOps.launchDma,
//       // ... arguments);
//  
//   if (!isPushSuccess)
//       goto cleanup;
//
// Continue to use CE HAL methods as-is, just omit the first two arugments:
// 1. unsigned **pbPut
// 2. unsigned *pbEnd
//
#define UVM_PUSH_METHOD(ret, pb, func, ...) \
    do { \
        NvU64 numMethods = func( \
                (unsigned**)&pb->pbOffset, \
                (unsigned*)((NvU64)pushbuffer->cpuBegin + \
                            UVM_PUSHBUFFER_RESERVATION_SIZE), \
                __VA_ARGS__); \
        if (numMethods) \
        { \
            pb->curOffset += numMethods; \
            ret = true; \
        } \
        else \
        { \
            ret = false; \
        } \
    } while(0)

//
// Channel Mgmt API Common Layer
//
// This level provides functionalities common to both Linux and Windows.
//
// 1. Start-up/Shut-down channel mgmt api.
// 2. Get/Submit pushbuffers.
// 3. Track/Wait on submitted work.
//
// uvm_* APIs are external APIs that the rest of UVM uses.
//
// uvm_channel_manager_* APIs are intenral APIs that only channel API should access.
//

//
// Channel Mgmt API Locking
// 
// The goal is to make locking in this API orthogonal to the rest of UVM.
// Taking locks elsewhere in other UVM components shouldn't affect locking
// inside the channel mgmt API.
//
// Lock abstraction:
//
//  uvm_lock_channel_manager
//  uvm_unlock_channel_manager
//
// Above functions are provided to lock/unlock UvmChannelManager.
//
// uvm_* functions use the above APIs, thus the caller doesn't need to
// grab locks before calling them.
//
// uvm_channel_manager* functions do not acquire/release
// locks. Caller is solely responsible to ensure proper locking before
// calling them.
//
// The only exception is: uvm_wait_for_tracker.
// * No lock needed to call this function.
//

// Forward declarations:
typedef struct UvmTrackerItem_tag     UvmTrackerItem;
typedef struct UvmTracker_tag         UvmTracker;
typedef struct UvmPushbuffer_tag      UvmPushbuffer;
typedef struct UvmRingbuffer_tag      UvmRingbuffer;
typedef struct UvmRingbufferPool_tag  UvmRingbufferPool;
typedef struct UvmChannel_tag         UvmChannel;
typedef struct UvmChannelTracking_tag UvmChannelTracking;
typedef struct UvmChannelPool_tag     UvmChannelPool;
typedef struct UvmChannelManager_tag  UvmChannelManager;

struct UvmTrackerItem_tag
{
    NvU64 seqNum;
    UvmChannel *channel;
};

struct UvmTracker_tag
{
    // TODO: add support for multiple tracker items on Linux.
    UvmTrackerItem item;
};

//
// UVM Pushbuffer Structure
//
// Pushbuffer structure is either owned by:
//   1. the pbFreeListHead in UvmRingbufferPool, or
//   2. the pbListHead in UvmRingbuffer.
//
// Use pbList to traverse the list that owns this pushbuffer structure.
//
struct UvmPushbuffer_tag
{
    //
    // Pushbuffer address and offset
    //
    // 1. CPU begin address of this pb segment
    // 2. GPU begin address of this pb segment
    // 3. The address which you can write methods to
    // 4. The current pushbuffer offset from the beginning
    //
    void         *cpuBegin;
    UvmGpuPointer gpuBegin;
    void         *pbOffset;
    NvU64         curOffset;

    UvmRingbuffer *ringbuffer;
    UvmChannel    *channel;

    NvU32 gpFifoOffset;
    NvU64 seqNum;

#if defined (__linux__)
    struct list_head pbList; 
    struct list_head pbSubmittedList;   // Connected to other pb pending completion
                                        // in the same channel.
#else
    LIST_ENTRY pbList;
#endif
};

//
// UVM Ringbuffer Structure
//
struct UvmRingbuffer_tag
{
    //
    // Ringbuffer address and offset
    //
    void         *cpuBegin;         // inclusive
    void         *cpuEnd;           // exclusive
    UvmGpuPointer gpuBegin;
    NvU64         curOffset;        // where we can start the next pushbuffer

#if defined (__linux__)
    struct list_head pbListHead;
    struct list_head ringbufferList;
#else
    LIST_ENTRY pbListHead;          // pb that are currently in-use.
    LIST_ENTRY ringbufferList;
#endif
};

struct UvmRingbufferPool_tag
{
    // Begin addresses of ringbuffer pool memory area
    UvmGpuPointer gpuPtr;
    void         *cpuPtr;

    unsigned      numRingbuffers;
#if defined (__linux__)
    struct list_head ringbufferFreeListHead;
    struct list_head ringbufferBusyListHead;
    struct list_head pbFreeListHead;    // free pb structures
#else
    LIST_ENTRY ringbufferListHead;
#endif
};

struct UvmChannelTracking_tag
{
    NvU32 curGpFifoOffset;
    NvU32 numReservedGpFifoEntries;
    NvU32 numFreeGpFifoEntries;

    UvmGpuPointer semaGpuPointer;
    void         *semaCpuPointer;

    NvU64 numCompletions;
    NvU64 numSubmissions;
};

struct UvmChannel_tag
{
    uvmGpuChannelHandle    hChannel;
    UvmGpuChannelPointers  channelInfo;
    uvmGpuCopyEngineHandle hCopyEngine;
    unsigned               ceClassNumber;
    UvmCopyOps             ceOps;

    // Channel status tracking information
    UvmChannelTracking trackingInfo;

#if defined (__linux__)
    struct list_head pbSubmittedListHead;       // List of pbs pending completion
                                                // in this channel.
    struct list_head channelList;
#else
    LIST_ENTRY channelList;
#endif
};

struct UvmChannelPool_tag
{
    uvmGpuSessionHandle      hSession;
    uvmGpuAddressSpaceHandle hVaSpace;
    UvmGpuCaps               gpuCaps;

    unsigned numChannels;
#if defined (__linux__)
    struct list_head    channelListHead;
#else
    LIST_ENTRY channelListHead;
#endif
};

struct UvmChannelManager_tag
{
    UvmChannelPool    channelPool;
    UvmRingbufferPool ringbufferPool;

#if defined (__linux__)
    struct rw_semaphore channelManagerLock;
#else
    FAST_MUTEX channelManagerLock;
#endif
};

//
// Initalize channel API global resources.
//
// Thread Safety: must be called only in a single thread.
//
RM_STATUS uvm_initialize_channel_mgmt_api(void);

//
// Deinitalize/destroy channel API global resources.
//
// Thread Safety: must be called only in a single thread.
//
void uvm_deinitialize_channel_mgmt_api(void);

//
// Create channel/pushbuffer resources associated with this GPU.
//
RM_STATUS uvm_create_channel_manager(UvmGpuUuid *gpuUuid,
                                     UvmChannelManager **channelManager);

//
// Destroy channel/pushbuffer resources tied to this channel manager.
//
// Caller must ensure proper sychronization.
//
// i.e. No one else should be touching this object during and after
//      calling this function.
//
void uvm_destroy_channel_manager(UvmChannelManager *channelManager);

//
// Lock UvmChannelManager
//
// This function should be called before using any fields in UvmChannelManager.
//
void uvm_lock_channel_manager(UvmChannelManager *channelManager);

//
// Unlock UvmChannelManager
//
// This function should be called after using UvmChannelManager
//
void uvm_unlock_channel_manager(UvmChannelManager *channelManager);

//
// Grab a pushbuffer from this channel manager.
//
// Lock channel manager before calling.
//
RM_STATUS uvm_get_pushbuffer(
        UvmChannelManager *channelManager, UvmPushbuffer **pushbuffer);

//
// Cancel a pushbuffer in this channel manager.
//
// Lock channel manager before calling.
//
void uvm_cancel_pushbuffer(
        UvmChannelManager *channelManager, UvmPushbuffer *pushbuffer);

//
// Submit this pushbuffer.
//
// Lock channel manager before calling.
//
RM_STATUS uvm_submit_pushbuffer(
        UvmChannelManager *channelManager, UvmPushbuffer *pushbuffer,
        UvmTracker *trackerToWait, UvmTracker *newTracker);

//
// Wait for tracker on the cpu side
//
// No lock required for this call.
//
RM_STATUS uvm_wait_for_tracker(UvmTracker *tracker);

//
// Uvm Channel Mgmt Internal APIs
//
// Hold a lock to UvmChannelManager before calling any of the below.
//

//
// Find a freely available pushbuffer structure for this particular
// channel manager. If we can't find any, allocate a new pushbuffer
// structure for this particular channel manager.
//
RM_STATUS uvm_channel_manager_alloc_pushbuffer_structure(
        UvmChannelManager *channelManager, UvmPushbuffer **pushbuffer);

//
// Free this pushbuffer structure into the freelist.
//
// This does not actually free the memory.
//
void uvm_channel_manager_free_pushbuffer_structure(
        UvmChannelManager *channelManager, UvmPushbuffer *pushbuffer);

//
// Get an available ringbuffer from the pool.
//
RM_STATUS uvm_channel_manager_get_ringbuffer(
        UvmChannelManager *channelManager, UvmRingbuffer **ringbuffer);

//
// Put a ringbuffer back to the pol's free list.
//
void uvm_channel_manager_put_ringbuffer(
        UvmChannelManager *channelManager, UvmRingbuffer *ringbuffer);

//
// Insert the pushbuffer into the ringbuffer's pb list.
//
void uvm_channel_manager_associate_pushbuffer_to_ringbuffer(
        UvmRingbuffer *ringbuffer, UvmPushbuffer *pushbuffer);

//
// Get a channel with spare GPFIFOs from the pool.
// Also reserve a GPFIFO entry
//
RM_STATUS uvm_channel_manager_get_channel(
        UvmChannelManager *channelManager, UvmChannel **channel);

//
// Call this function when you cancel a pushbuffer, and want to
// restore the channel to its previous state.
//
// Put a channel back to the beginning the pool.
// Also unreserve a GPFIFO entry.
//
void uvm_channel_manager_put_channel(
        UvmChannelManager *channelManager, UvmChannel *channel);

//
// Call this function when you submit a pushbuffer, and want to
// update the book-keeping stuff in the channel.
//
// 1. Insert the pushbuffer into the ringbuffer's pb submitted list.
// 2. Update channel book-keeping to reflect a GPFIFO-entry is
//    currently pending completion
// 3. Update channel progress info.
//
void uvm_channel_manager_associate_pushbuffer_to_channel(
        UvmChannel *channel, UvmPushbuffer *pushbuffer);

//
// Update all channels progress information.
//
void uvm_channel_manager_update_all_channel_progress(
        UvmChannelManager *channelManager);

//
// Update channel progress information.
//
void uvm_channel_manager_update_channel_progress(UvmChannel *channel);

//
// Reclaim pushbuffer memory and GPFIFO entries in this channel manager.
//
RM_STATUS uvm_channel_manager_reclaim(UvmChannelManager *channelManager);

#endif // _UVM_CHANNEL_MGMT_COMMON_H_
